import pytest

from pullkin import AioPullkin, Pullkin


@pytest.mark.asyncio
async def test_aio_registration():
    client = AioPullkin()
    credentials = client.register(581003993230)


def test_registration():
    client = Pullkin()
    credentials = client.register(581003993230)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: mcs.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import List

import betterproto


class ClientEventType(betterproto.Enum):
    UNKNOWN = 0
    DISCARDED_EVENTS = 1
    FAILED_CONNECTION = 2
    SUCCESSFUL_CONNECTION = 3


class LoginRequestAuthService(betterproto.Enum):
    ANDROID_ID = 2


class IqStanzaIqType(betterproto.Enum):
    GET = 0
    SET = 1
    RESULT = 2
    IQ_ERROR = 3


@dataclass
class HeartbeatPing(betterproto.Message):
    """*TAG: 0"""

    stream_id: int = betterproto.int32_field(1)
    last_stream_id_received: int = betterproto.int32_field(2)
    status: int = betterproto.int64_field(3)


@dataclass
class HeartbeatAck(betterproto.Message):
    """*TAG: 1"""

    stream_id: int = betterproto.int32_field(1)
    last_stream_id_received: int = betterproto.int32_field(2)
    status: int = betterproto.int64_field(3)


@dataclass
class ErrorInfo(betterproto.Message):
    code: int = betterproto.int32_field(1)
    message: str = betterproto.string_field(2)
    type: str = betterproto.string_field(3)
    extension: "Extension" = betterproto.message_field(4)


@dataclass
class Setting(betterproto.Message):
    name: str = betterproto.string_field(1)
    value: str = betterproto.string_field(2)


@dataclass
class HeartbeatStat(betterproto.Message):
    ip: str = betterproto.string_field(1)
    timeout: bool = betterproto.bool_field(2)
    interval_ms: int = betterproto.int32_field(3)


@dataclass
class HeartbeatConfig(betterproto.Message):
    upload_stat: bool = betterproto.bool_field(1)
    ip: str = betterproto.string_field(2)
    interval_ms: int = betterproto.int32_field(3)


@dataclass
class ClientEvent(betterproto.Message):
    """
    ClientEvents are used to inform the server of failed and successful
    connections.
    """

    # Common fields [1-99]
    type: "ClientEventType" = betterproto.enum_field(1)
    # Fields for DISCARDED_EVENTS messages [100-199]
    number_discarded_events: int = betterproto.uint32_field(100)
    # Fields for FAILED_CONNECTION and SUCCESSFUL_CONNECTION messages [200-299]
    # Network type is a value in net::NetworkChangeNotifier::ConnectionType.
    network_type: int = betterproto.int32_field(200)
    time_connection_started_ms: int = betterproto.uint64_field(202)
    time_connection_ended_ms: int = betterproto.uint64_field(203)
    # Error code should be a net::Error value.
    error_code: int = betterproto.int32_field(204)
    # Fields for SUCCESSFUL_CONNECTION messages [300-399]
    time_connection_established_ms: int = betterproto.uint64_field(300)


@dataclass
class LoginRequest(betterproto.Message):
    """*TAG: 2"""

    id: str = betterproto.string_field(1)
    # string. mcs.android.com.
    domain: str = betterproto.string_field(2)
    # Decimal android ID
    user: str = betterproto.string_field(3)
    resource: str = betterproto.string_field(4)
    # Secret
    auth_token: str = betterproto.string_field(5)
    # Format is: android-HEX_DEVICE_ID The user is the decimal value.
    device_id: str = betterproto.string_field(6)
    # RMQ1 - no longer used
    last_rmq_id: int = betterproto.int64_field(7)
    setting: List["Setting"] = betterproto.message_field(8)
    # optional int32 compress = 9;
    received_persistent_id: List[str] = betterproto.string_field(10)
    adaptive_heartbeat: bool = betterproto.bool_field(12)
    heartbeat_stat: "HeartbeatStat" = betterproto.message_field(13)
    # Must be true.
    use_rmq2: bool = betterproto.bool_field(14)
    account_id: int = betterproto.int64_field(15)
    # ANDROID_ID = 2
    auth_service: "LoginRequestAuthService" = betterproto.enum_field(16)
    network_type: int = betterproto.int32_field(17)
    status: int = betterproto.int64_field(18)
    # Events recorded on the client after the last successful connection.
    client_event: List["ClientEvent"] = betterproto.message_field(22)


@dataclass
class LoginResponse(betterproto.Message):
    """* TAG: 3"""

    id: str = betterproto.string_field(1)
    # Not used.
    jid: str = betterproto.string_field(2)
    # Null if login was ok.
    error: "ErrorInfo" = betterproto.message_field(3)
    setting: List["Setting"] = betterproto.message_field(4)
    stream_id: int = betterproto.int32_field(5)
    # Should be "1"
    last_stream_id_received: int = betterproto.int32_field(6)
    heartbeat_config: "HeartbeatConfig" = betterproto.message_field(7)
    # used by the client to synchronize with the server timestamp.
    server_timestamp: int = betterproto.int64_field(8)


@dataclass
class StreamErrorStanza(betterproto.Message):
    type: str = betterproto.string_field(1)
    text: str = betterproto.string_field(2)


@dataclass
class Close(betterproto.Message):
    """* TAG: 4"""

    pass


@dataclass
class Extension(betterproto.Message):
    # 12: SelectiveAck 13: StreamAck
    id: int = betterproto.int32_field(1)
    data: bytes = betterproto.bytes_field(2)


@dataclass
class IqStanza(betterproto.Message):
    """
    * TAG: 7 IqRequest must contain a single extension.  IqResponse may contain
    0 or 1 extensions.
    """

    rmq_id: int = betterproto.int64_field(1)
    type: "IqStanzaIqType" = betterproto.enum_field(2)
    id: str = betterproto.string_field(3)
    from_: str = betterproto.string_field(4)
    to: str = betterproto.string_field(5)
    error: "ErrorInfo" = betterproto.message_field(6)
    # Only field used in the 38+ protocol (besides common
    # last_stream_id_received, status, rmq_id)
    extension: "Extension" = betterproto.message_field(7)
    persistent_id: str = betterproto.string_field(8)
    stream_id: int = betterproto.int32_field(9)
    last_stream_id_received: int = betterproto.int32_field(10)
    account_id: int = betterproto.int64_field(11)
    status: int = betterproto.int64_field(12)


@dataclass
class AppData(betterproto.Message):
    key: str = betterproto.string_field(1)
    value: str = betterproto.string_field(2)


@dataclass
class DataMessageStanza(betterproto.Message):
    """* TAG: 8"""

    # This is the message ID, set by client, DMP.9 (message_id)
    id: str = betterproto.string_field(2)
    # Project ID of the sender, DMP.1
    from_: str = betterproto.string_field(3)
    # Part of DMRequest - also the key in DataMessageProto.
    to: str = betterproto.string_field(4)
    # Package name. DMP.2
    category: str = betterproto.string_field(5)
    # The collapsed key, DMP.3
    token: str = betterproto.string_field(6)
    # User data + GOOGLE. prefixed special entries, DMP.4
    app_data: List["AppData"] = betterproto.message_field(7)
    # Not used.
    from_trusted_server: bool = betterproto.bool_field(8)
    # Part of the ACK protocol, returned in DataMessageResponse on server side.
    # It's part of the key of DMP.
    persistent_id: str = betterproto.string_field(9)
    # In-stream ack. Increments on each message sent - a bit redundant Not used
    # in DMP/DMR.
    stream_id: int = betterproto.int32_field(10)
    last_stream_id_received: int = betterproto.int32_field(11)
    # Sent by the device shortly after registration.
    reg_id: str = betterproto.string_field(13)
    # serial number of the target user, DMP.8 It is the 'serial number' according
    # to user manager.
    device_user_id: int = betterproto.int64_field(16)
    # Time to live, in seconds.
    ttl: int = betterproto.int32_field(17)
    # Timestamp ( according to client ) when message was sent by app, in seconds
    sent: int = betterproto.int64_field(18)
    # How long has the message been queued before the flush, in seconds. This is
    # needed to account for the time difference between server and client: server
    # should adjust 'sent' based on its 'receive' time.
    queued: int = betterproto.int32_field(19)
    status: int = betterproto.int64_field(20)
    # Optional field containing the binary payload of the message.
    raw_data: bytes = betterproto.bytes_field(21)
    # If set the server requests immediate ack. Used for important messages and
    # for testing.
    immediate_ack: bool = betterproto.bool_field(24)


@dataclass
class StreamAck(betterproto.Message):
    """
    *Included in IQ with ID 13, sent from client or server after 10
    unconfirmedmessages.
    """

    pass


@dataclass
class SelectiveAck(betterproto.Message):
    """*Included in IQ sent after LoginResponse from server with ID 12."""

    id: List[str] = betterproto.string_field(1)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: android_checkin.proto
# plugin: python-betterproto
from dataclasses import dataclass

import betterproto


class DeviceType(betterproto.Enum):
    """
    enum values correspond to the type of device. Used in the
    AndroidCheckinProto and Device proto.
    """

    # Android Device
    DEVICE_ANDROID_OS = 1
    # Apple IOS device
    DEVICE_IOS_OS = 2
    # Chrome browser - Not Chrome OS.  No hardware records.
    DEVICE_CHROME_BROWSER = 3
    # Chrome OS
    DEVICE_CHROME_OS = 4


class ChromeBuildProtoPlatform(betterproto.Enum):
    PLATFORM_WIN = 1
    PLATFORM_MAC = 2
    PLATFORM_LINUX = 3
    PLATFORM_CROS = 4
    PLATFORM_IOS = 5
    PLATFORM_ANDROID = 6


class ChromeBuildProtoChannel(betterproto.Enum):
    CHANNEL_STABLE = 1
    CHANNEL_BETA = 2
    CHANNEL_DEV = 3
    CHANNEL_CANARY = 4
    CHANNEL_UNKNOWN = 5


@dataclass
class ChromeBuildProto(betterproto.Message):
    """Build characteristics unique to the Chrome browser, and Chrome OS"""

    # The platform of the device.
    platform: "ChromeBuildProtoPlatform" = betterproto.enum_field(1)
    # The Chrome instance's version.
    chrome_version: str = betterproto.string_field(2)
    # The Channel (build type) of Chrome.
    channel: "ChromeBuildProtoChannel" = betterproto.enum_field(3)


@dataclass
class AndroidCheckinProto(betterproto.Message):
    """Information sent by the device in a "checkin" request."""

    # Miliseconds since the Unix epoch of the device's last successful checkin.
    last_checkin_msec: int = betterproto.int64_field(2)
    # The current MCC+MNC of the mobile device's current cell.
    cell_operator: str = betterproto.string_field(6)
    # The MCC+MNC of the SIM card (different from operator if the device is
    # roaming, for instance).
    sim_operator: str = betterproto.string_field(7)
    # The device's current roaming state (reported starting in eclair builds).
    # Currently one of "{,not}mobile-{,not}roaming", if it is present at all.
    roaming: str = betterproto.string_field(8)
    # For devices supporting multiple user profiles (which may be supported
    # starting in jellybean), the ordinal number of the profile that is checking
    # in.  This is 0 for the primary profile (which can't be changed without
    # wiping the device), and 1,2,3,... for additional profiles (which can be
    # added and deleted freely).
    user_number: int = betterproto.int32_field(9)
    # Class of device.  Indicates the type of build proto
    # (IosBuildProto/ChromeBuildProto/AndroidBuildProto) That is included in this
    # proto
    type: "DeviceType" = betterproto.enum_field(12)
    # For devices running MCS on Chrome, build-specific characteristics of the
    # browser.  There are no hardware aspects (except for ChromeOS). This will
    # only be populated for Chrome builds/ChromeOS devices
    chrome_build: "ChromeBuildProto" = betterproto.message_field(13)



from copy import deepcopy
from typing import Optional


class NotificationData:
    def __init__(self, data: dict = None):
        if not data:
            data = {}
        self.raw_data: dict = deepcopy(data)
        self.title: str = data.pop("title", None)
        self.body: str = data.pop("body", None)
        if data:
            for k, v in data.items():
                self.__setattr__(k, v)

    def __str__(self):
        return str(self.__dict__)


class Message:
    def __init__(self, data: dict = None):
        if not data:
            data = {}
        self.raw_data: dict = deepcopy(data)
        self.sender_id: Optional[str] = data.pop("from", None)
        self.priority: Optional[str] = data.pop("priority", None)
        self.notification: NotificationData = NotificationData(
            data.pop("notification", None)
        )
        self.fcmMessageId: Optional[str] = data.pop("fcmMessageId", None)
        if data:
            for k, v in data.items():
                self.__setattr__(k, v)

    def __getitem__(self, item):
        return self.raw_data[item]

    def __str__(self):
        return str(self.__dict__)



import asyncio
import inspect
import json
import struct
import traceback
from asyncio import StreamReader, StreamWriter
from base64 import urlsafe_b64decode
from binascii import hexlify
from typing import Callable, Optional, Union, AsyncGenerator

import cryptography.hazmat.primitives.serialization as serialization
import http_ece
from cryptography.hazmat.backends import default_backend
from loguru import logger

from pullkin.client_base import PullkinBase
from pullkin.proto.mcs_proto import *
from pullkin.models.message import Message

logger.disable("pullkin")


class AioPullkin(PullkinBase):
    def __init__(self):
        super().__init__()
        self.__reader: Optional[StreamReader] = None
        self.__writer: Optional[StreamWriter] = None
        self.credentials: dict = {}
        self.persistent_ids: list = []
        self.callback: Callable = None
        self.on_notification_handlers: list = []
        self.once = True

    def on_notification(self, filter: Callable = lambda *a, **k: True):
        """
        Decorator

        Registers a new callback with filter rule to trigger this callback

        First param to be passed into filter or callback: some object. Now is empty dict.
        Later you will be able to passed your object
        Second param: Notification object. You can read about Notification here: TODO: add a link
        Third param: Message object that get from socket. With all data about message. Read hereL TODO: add a link

        :: python code
            from pullkin import AioPullkin, Notification, DataMessageStanza as Message
            pullkin = AuiPullkin()

            @pullkin.on_notification(lambda o, n, m: n.title == "Moi koleni zamerzli")
            def on_zemfira(obj, notification, message):
                print(notification.notification.body)

            @pullkin.on_notification(lambda o, n, m: n.priority == "normal")
            async def on_normal(obj: dict, notification: Notification, message: Message):
                print(f"#{message.id} @{message.sent})
                print(notification")
                await asyncio.sleep(2)
                print("=-)")

        """

        def decorator(callback):
            self.on_notification_handlers.append(
                {"callback": callback, "filter": filter}
            )
            return callback

        return decorator

    def register_on_notification_handler(
        self, filter: Callable = None, callback: Callable = None
    ):
        """
        Function

        Registers a new callback with filter rule to trigger this callback

        filter: rule for calling thit callback
        callback: function to be called if the filter-function returns True

        First param to be passed into filter and callback: some object. Now is empty dict.
        Later you will be able to passed your object
        Second param: Notification object. You can read about Notification here: TODO: add a link
        Third param: Message object that get from socket. With all data about message. Read hereL TODO: add a link

        :: python code
            from pullkin import AioPullkin, Notification, DataMessageStanza as Message
            pullkin = AuiPullkin()

            @pullkin.on_notification(lambda o, n, m: n.title == "Moi koleni zamerzli")
            def on_zemfira(obj, notification, message):
                print(notification.notification.body)

            @pullkin.on_notification(lambda o, n, m: n.priority == "normal")
            async def on_normal(obj: dict, notification: Notification, message: Message):
                print(f"#{message.id} @{message.sent})
                print(notification")
                await asyncio.sleep(2)
                print("=-)")

        """
        self.on_notification_handlers.append({"callback": callback, "filter": filter})

    async def __run_on_notification_callbacks(self, obj, notification, data_message):
        x = 0
        for handler in self.on_notification_handlers:
            if handler["filter"](obj, notification, data_message):
                if inspect.iscoroutinefunction(handler["callback"]):
                    await handler["callback"](obj, notification, data_message)
                else:
                    handler["callback"](obj, notification, data_message)
                x += 1
                if self.once:
                    break
        else:
            if self.callback:
                if inspect.iscoroutinefunction(self.callback):
                    await self.callback(obj, notification, data_message)
                else:
                    self.callback(obj, notification, data_message)

        if not x:
            logger.debug('No one callback was called')

    @classmethod
    async def aioregister(cls, sender_id):
        """
        Async version. Register "app" for receive pushed

        Returns "app"-credential in dict for receive "personal" push by token
        """
        return await cls._register(sender_id)

    async def __open_connection(self) -> None:
        import ssl

        ssl_ctx = ssl.create_default_context()
        self.__reader, self.__writer = await asyncio.open_connection(
            self.PUSH_HOST, self.PUSH_PORT, ssl=ssl_ctx
        )
        logger.debug(
            f"Connected to SSL socket {self.PUSH_HOST}:{self.PUSH_PORT} with default ssl_context"
        )

    async def __aioread(self, size) -> bytes:
        buf = b""
        while len(buf) < size:
            buf += await self.__reader.read(size - len(buf))
        return buf

    async def __aioread_varint32(self) -> int:
        res = 0
        shift = 0
        while True:
            (b,) = struct.unpack("B", await self.__aioread(1))
            res |= (b & 0x7F) << shift
            if (b & 0x80) == 0:
                break
            shift += 7
        return res

    async def __aiosend(self, packet) -> None:
        logger.debug(f"Send")
        header = bytearray([self.MCS_VERSION, self.PACKET_BY_TAG.index(type(packet))])
        logger.debug(f"Packet:\n'{packet}'")
        payload = packet.SerializeToString()
        buf = bytes(header) + self._encode_varint32(len(payload)) + payload
        logger.debug(f"HEX buffer:\n`{hexlify(buf)}`")
        self.__writer.write(buf)
        await self.__writer.drain()

    async def __aiorecv(self, first=False) -> Optional[PullkinBase.packet_union]:
        logger.debug(f"Receive")
        if first:
            version, tag = struct.unpack("BB", await self.__aioread(2))
            logger.debug(f"Version {version}")
            if version < self.MCS_VERSION and version != 38:
                raise RuntimeError(f"Protocol version {version} unsupported")
        else:
            (tag,) = struct.unpack("B", await self.__aioread(1))
        logger.debug(f"Tag {tag} ({self.PACKET_BY_TAG[tag]})")
        size = await self.__aioread_varint32()
        logger.debug(f"Size {size}")
        if size >= 0:
            buf = await self.__aioread(size)
            logger.debug(f"HEX buffer:\n`{hexlify(buf)}`")
            packet_class = self.PACKET_BY_TAG[tag]
            payload = packet_class()
            payload.parse(buf)
            logger.debug(f"Payload:\n`{payload}`")
            return payload
        return None

    async def __aiolisten_once(
        self,
    ) -> None:
        load_der_private_key = serialization.load_der_private_key

        p = await self.__aiorecv()
        if type(p) is not DataMessageStanza:
            return
        if self._is_deleted_message(p):
            return
        crypto_key = self._app_data_by_key(p, "crypto-key", False)
        salt = self._app_data_by_key(p, "encryption", False)
        logger.debug(f"crypto-key: {crypto_key}, salt: {salt}")
        if not (salt and crypto_key):
            return
        crypto_key = crypto_key[3:]  # strip dh=
        salt = salt[5:]  # strip salt=
        crypto_key = urlsafe_b64decode(crypto_key.encode("ascii"))
        salt = urlsafe_b64decode(salt.encode("ascii"))
        der_data = self.credentials["keys"]["private"]
        der_data = urlsafe_b64decode(der_data.encode("ascii") + b"========")
        secret = self.credentials["keys"]["secret"]
        secret = urlsafe_b64decode(secret.encode("ascii") + b"========")
        privkey = load_der_private_key(
            der_data, password=None, backend=default_backend()
        )
        decrypted = http_ece.decrypt(
            p.raw_data,
            salt=salt,
            private_key=privkey,
            dh=crypto_key,
            version="aesgcm",
            auth_secret=secret,
        )
        notification = Message(json.loads(decrypted.decode("utf-8")))
        await self.__run_on_notification_callbacks({}, notification, p)

    async def __aiolisten_start(self) -> None:
        await self.gcm_check_in(**self.credentials["gcm"])
        req = LoginRequest()
        req.adaptive_heartbeat = False
        req.auth_service = 2
        req.auth_token = self.credentials["gcm"]["securityToken"]
        req.id = "chrome-91.0.3234.0"
        req.domain = "mcs.android.com"
        req.device_id = "android-%x" % int(self.credentials["gcm"]["androidId"])
        req.network_type = 1
        req.resource = self.credentials["gcm"]["androidId"]
        req.user = self.credentials["gcm"]["androidId"]
        req.use_rmq2 = True
        req.setting.append(Setting(name="new_vc", value="1"))
        req.received_persistent_id.extend(self.persistent_ids)
        await self.__aiosend(req)
        await self.__aiorecv(first=True)

    async def __aiolisten_coroutine(self) -> AsyncGenerator:
        while True:
            yield await self.__aiolisten_once()

    async def listen_coroutine(self) -> AsyncGenerator:
        """
        Return a listener-coroutine

        Every coroutine iteration is one received push
        (or not, because it starts reading and waiting for data on socket)

        You can use coroutine like this, for example:

        :: python code
            # <some code>

            await coroutine.asend(None)  # One push will be received and distributed

            # <some code>

            for _ in range(10): # Ten pushes will be received
                await coroutine.asend(None)

            # <some code>

        :return: coroutine
        """
        if not (self.__reader or self.__writer):
            await self.__open_connection()
        await self.__aiolisten_start()
        coroutine = self.__aiolisten_coroutine()
        return coroutine

    async def listen_forever(self, timer: Union[int, float] = 0.05) -> None:
        """
        Listens for push notifications

        Runs an endless loop for reading notifications and distributing among callbacks based on filter results

        :param timer: timer in seconds between receive iteration
        :type timer: ``int`` or ``float``, optional, default ``0.05``
        """
        if not (self.__reader or self.__writer):
            await self.__open_connection()

        await self.__aiolisten_start()
        coroutine = self.__aiolisten_coroutine()
        try:
            while self.__reader and self.__writer:
                await coroutine.asend(None)
                await asyncio.sleep(timer)
        except Exception as e:
            print(traceback.format_exc())
        finally:
            if self.__writer:
                self.__writer.close()
                await self.__writer.wait_closed()

    async def close(self):
        try:
            if self._http_client:
                await self._http_client.aclose()
                self._http_client = None
            if self.__writer:
                self.__writer.close()
                await self.__writer.wait_closed()
            self.__writer = None
            self.__reader = None
        except ConnectionResetError:
            return



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: android_checkin.proto, checkin.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import List

import betterproto


class DeviceType(betterproto.Enum):
    """
    enum values correspond to the type of device. Used in the
    AndroidCheckinProto and Device proto.
    """

    # Android Device
    DEVICE_ANDROID_OS = 1
    # Apple IOS device
    DEVICE_IOS_OS = 2
    # Chrome browser - Not Chrome OS.  No hardware records.
    DEVICE_CHROME_BROWSER = 3
    # Chrome OS
    DEVICE_CHROME_OS = 4


class ChromeBuildProtoPlatform(betterproto.Enum):
    PLATFORM_WIN = 1
    PLATFORM_MAC = 2
    PLATFORM_LINUX = 3
    PLATFORM_CROS = 4
    PLATFORM_IOS = 5
    PLATFORM_ANDROID = 6


class ChromeBuildProtoChannel(betterproto.Enum):
    CHANNEL_STABLE = 1
    CHANNEL_BETA = 2
    CHANNEL_DEV = 3
    CHANNEL_CANARY = 4
    CHANNEL_UNKNOWN = 5


@dataclass
class ChromeBuildProto(betterproto.Message):
    """Build characteristics unique to the Chrome browser, and Chrome OS"""

    # The platform of the device.
    platform: "ChromeBuildProtoPlatform" = betterproto.enum_field(1)
    # The Chrome instance's version.
    chrome_version: str = betterproto.string_field(2)
    # The Channel (build type) of Chrome.
    channel: "ChromeBuildProtoChannel" = betterproto.enum_field(3)


@dataclass
class AndroidCheckinProto(betterproto.Message):
    """Information sent by the device in a "checkin" request."""

    # Miliseconds since the Unix epoch of the device's last successful checkin.
    last_checkin_msec: int = betterproto.int64_field(2)
    # The current MCC+MNC of the mobile device's current cell.
    cell_operator: str = betterproto.string_field(6)
    # The MCC+MNC of the SIM card (different from operator if the device is
    # roaming, for instance).
    sim_operator: str = betterproto.string_field(7)
    # The device's current roaming state (reported starting in eclair builds).
    # Currently one of "{,not}mobile-{,not}roaming", if it is present at all.
    roaming: str = betterproto.string_field(8)
    # For devices supporting multiple user profiles (which may be supported
    # starting in jellybean), the ordinal number of the profile that is checking
    # in.  This is 0 for the primary profile (which can't be changed without
    # wiping the device), and 1,2,3,... for additional profiles (which can be
    # added and deleted freely).
    user_number: int = betterproto.int32_field(9)
    # Class of device.  Indicates the type of build proto
    # (IosBuildProto/ChromeBuildProto/AndroidBuildProto) That is included in this
    # proto
    type: "DeviceType" = betterproto.enum_field(12)
    # For devices running MCS on Chrome, build-specific characteristics of the
    # browser.  There are no hardware aspects (except for ChromeOS). This will
    # only be populated for Chrome builds/ChromeOS devices
    chrome_build: "ChromeBuildProto" = betterproto.message_field(13)


@dataclass
class GservicesSetting(betterproto.Message):
    """A concrete name/value pair sent to the device's Gservices database."""

    name: bytes = betterproto.bytes_field(1)
    value: bytes = betterproto.bytes_field(2)


@dataclass
class AndroidCheckinRequest(betterproto.Message):
    """Devices send this every few hours to tell us how they're doing."""

    # IMEI (used by GSM phones) is sent and stored as 15 decimal digits; the 15th
    # is a check digit.
    imei: str = betterproto.string_field(1)
    # MEID (used by CDMA phones) is sent and stored as 14 hexadecimal digits (no
    # check digit).
    meid: str = betterproto.string_field(10)
    # MAC address (used by non-phone devices).  12 hexadecimal digits; no
    # separators (eg "0016E6513AC2", not "00:16:E6:51:3A:C2").
    mac_addr: List[str] = betterproto.string_field(9)
    # An array parallel to mac_addr, describing the type of interface. Currently
    # accepted values: "wifi", "ethernet", "bluetooth".  If not present, "wifi"
    # is assumed.
    mac_addr_type: List[str] = betterproto.string_field(19)
    # Serial number (a manufacturer-defined unique hardware identifier).
    # Alphanumeric, case-insensitive.
    serial_number: str = betterproto.string_field(16)
    # Older CDMA networks use an ESN (8 hex digits) instead of an MEID.
    esn: str = betterproto.string_field(17)
    id: int = betterproto.int64_field(2)
    logging_id: int = betterproto.int64_field(7)
    digest: str = betterproto.string_field(3)
    locale: str = betterproto.string_field(6)
    checkin: "AndroidCheckinProto" = betterproto.message_field(4)
    # DEPRECATED, see AndroidCheckinProto.requested_group
    desired_build: str = betterproto.string_field(5)
    # Blob of data from the Market app to be passed to Market API server
    market_checkin: str = betterproto.string_field(8)
    # SID cookies of any google accounts stored on the phone.  Not logged.
    account_cookie: List[str] = betterproto.string_field(11)
    # Time zone.  Not currently logged.
    time_zone: str = betterproto.string_field(12)
    # Security token used to validate the checkin request. Required for android
    # IDs issued to Froyo+ devices, not for legacy IDs.
    security_token: float = betterproto.fixed64_field(13)
    # Version of checkin protocol. There are currently two versions: - version
    # field missing: android IDs are assigned based on   hardware identifiers.
    # unsecured in the sense that you can   "unregister" someone's phone by
    # sending a registration request   with their IMEI/MEID/MAC. - version=2:
    # android IDs are assigned randomly.  The device is   sent a security token
    # that must be included in all future   checkins for that android id. -
    # version=3: same as version 2, but the 'fragment' field is   provided, and
    # the device understands incremental updates to the   gservices table (ie,
    # only returning the keys whose values have   changed.) (version=1 was
    # skipped to avoid confusion with the "missing" version field that is
    # effectively version 1.)
    version: int = betterproto.int32_field(14)
    # OTA certs accepted by device (base-64 SHA-1 of cert files).  Not logged.
    ota_cert: List[str] = betterproto.string_field(15)
    # A single CheckinTask on the device may lead to multiple checkin requests if
    # there is too much log data to upload in a single request.  For version 3
    # and up, this field will be filled in with the number of the request,
    # starting with 0.
    fragment: int = betterproto.int32_field(20)
    # For devices supporting multiple users, the name of the current profile
    # (they all check in independently, just as if they were multiple physical
    # devices).  This may not be set, even if the device is using multiuser.
    # (checkin.user_number should be set to the ordinal of the user.)
    user_name: str = betterproto.string_field(21)
    # For devices supporting multiple user profiles, the serial number for the
    # user checking in.  Not logged.  May not be set, even if the device
    # supportes multiuser.  checkin.user_number is the ordinal of the user (0, 1,
    # 2, ...), which may be reused if users are deleted and re-created.
    # user_serial_number is never reused (unless the device is wiped).
    user_serial_number: int = betterproto.int32_field(22)


@dataclass
class AndroidCheckinResponse(betterproto.Message):
    """The response to the device."""

    stats_ok: bool = betterproto.bool_field(1)
    time_msec: int = betterproto.int64_field(3)
    # Provisioning is sent if the request included an obsolete digest. For
    # version <= 2, 'digest' contains the digest that should be sent back to the
    # server on the next checkin, and 'setting' contains the entire gservices
    # table (which replaces the entire current table on the device). for version
    # >= 3, 'digest' will be absent.  If 'settings_diff' is false, then 'setting'
    # contains the entire table, as in version 2.  If 'settings_diff' is true,
    # then 'delete_setting' contains the keys to delete, and 'setting' contains
    # only keys to be added or for which the value has changed.  All other keys
    # in the current table should be left untouched.  If 'settings_diff' is
    # absent, don't touch the existing gservices table.
    digest: str = betterproto.string_field(4)
    settings_diff: bool = betterproto.bool_field(9)
    delete_setting: List[str] = betterproto.string_field(10)
    setting: List["GservicesSetting"] = betterproto.message_field(5)
    market_ok: bool = betterproto.bool_field(6)
    android_id: float = betterproto.fixed64_field(7)
    security_token: float = betterproto.fixed64_field(8)
    version_info: str = betterproto.string_field(11)



import asyncio
import inspect
import json
import struct
import time
from base64 import urlsafe_b64decode
from binascii import hexlify

from loguru import logger

from pullkin.client_base import PullkinBase
from pullkin.proto.mcs_proto import *

logger.disable("pullkin")


class Pullkin(PullkinBase):
    def __init__(self):
        super().__init__()
        ...

    def __read(self, s, size):
        buf = b""
        while len(buf) < size:
            buf += s.recv(size - len(buf))
        return buf

    def __read_varint32(self, s):
        res = 0
        shift = 0
        while True:
            (b,) = struct.unpack("B", self.__read(s, 1))
            res |= (b & 0x7F) << shift
            if (b & 0x80) == 0:
                break
            shift += 7
        return res

    def __send(self, s, packet):
        logger.debug(f"Send")
        header = bytearray([self.MCS_VERSION, self.PACKET_BY_TAG.index(type(packet))])
        logger.debug(f"Packet:\n`{packet}`")
        payload = packet.SerializeToString()
        buf = bytes(header) + self._encode_varint32(len(payload)) + payload
        logger.debug(f"HEX buffer:\n`{hexlify(buf)}`")
        n = len(buf)
        total = 0
        while total < n:
            sent = s.send(buf[total:])
            if sent == 0:
                raise RuntimeError("socket connection broken")
            total += sent

    def __recv(self, s, first=False):
        logger.debug(f"Receive")
        if first:
            version, tag = struct.unpack("BB", self.__read(s, 2))
            logger.debug(f"Version {version}")
            if version < self.MCS_VERSION and version != 38:
                raise RuntimeError("protocol version {} unsupported".format(version))
        else:
            (tag,) = struct.unpack("B", self.__read(s, 1))
        logger.debug(f"Tag {tag} ({self.PACKET_BY_TAG[tag]})")
        size = self.__read_varint32(s)
        logger.debug(f"Size {size}")
        if size >= 0:
            buf = self.__read(s, size)
            logger.debug(f"HEX buffer:\n`{hexlify(buf)}`")
            packet_class = self.PACKET_BY_TAG[tag]
            payload = packet_class()
            payload.parse(buf)
            logger.debug(f"Payload:\n`{payload}`")
            return payload
        return None

    def __listen(
        self, s, credentials, callback, persistent_ids, obj, timer=0, is_alive=True
    ):
        import cryptography.hazmat.primitives.serialization as serialization
        import http_ece
        from cryptography.hazmat.backends import default_backend

        load_der_private_key = serialization.load_der_private_key

        self.gcm_check_in(**credentials["gcm"])
        req = LoginRequest()
        req.adaptive_heartbeat = False
        req.auth_service = 2
        req.auth_token = credentials["gcm"]["securityToken"]
        req.id = "chrome-91.0.3234.0"
        req.domain = "mcs.android.com"
        req.device_id = "android-%x" % int(credentials["gcm"]["androidId"])
        req.network_type = 1
        req.resource = credentials["gcm"]["androidId"]
        req.user = credentials["gcm"]["androidId"]
        req.use_rmq2 = True
        req.setting.append(Setting(name="new_vc", value="1"))
        req.received_persistent_id.extend(persistent_ids)
        self.__send(s, req)
        self.__recv(s, first=True)
        while is_alive:
            p = self.__recv(s)
            if type(p) is not DataMessageStanza:
                continue
            crypto_key = self._app_data_by_key(p, "crypto-key")[3:]  # strip dh=
            salt = self._app_data_by_key(p, "encryption")[5:]  # strip salt=
            crypto_key = urlsafe_b64decode(crypto_key.encode("ascii"))
            salt = urlsafe_b64decode(salt.encode("ascii"))
            der_data = credentials["keys"]["private"]
            der_data = urlsafe_b64decode(der_data.encode("ascii") + b"========")
            secret = credentials["keys"]["secret"]
            secret = urlsafe_b64decode(secret.encode("ascii") + b"========")
            privkey = load_der_private_key(
                der_data, password=None, backend=default_backend()
            )
            decrypted = http_ece.decrypt(
                p.raw_data,
                salt=salt,
                private_key=privkey,
                dh=crypto_key,
                version="aesgcm",
                auth_secret=secret,
            )
            if inspect.iscoroutinefunction(callback):
                asyncio.run(callback(obj, json.loads(decrypted.decode("utf-8")), p))
            else:
                callback(obj, json.loads(decrypted.decode("utf-8")), p)
            if timer:
                time.sleep(timer)

    def listen(
        self,
        credentials,
        callback,
        received_persistent_ids=None,
        obj=None,
        timer=0,
        is_alive=True,
    ):
        """
        listens for push notifications

        credentials: credentials object returned by register()
        callback(obj, notification, data_message): called on notifications
        received_persistent_ids: any persistent id's you already received.
                                 array of strings
        obj: optional arbitrary value passed to callback
        """
        if received_persistent_ids is None:
            received_persistent_ids = []
        import socket
        import ssl

        host = "mtalk.google.com"
        ssl_ctx = ssl.create_default_context()
        sock = socket.create_connection((self.PUSH_HOST, self.PUSH_PORT))
        s = ssl_ctx.wrap_socket(sock, server_hostname=host)
        logger.debug(
            f"Connected to SSL socket {self.PUSH_HOST}:{self.PUSH_PORT} with default ssl_context"
        )
        self.__listen(
            s, credentials, callback, received_persistent_ids, obj, timer, is_alive
        )
        s.close()
        sock.close()



import asyncio
import json
import os
import time
from base64 import urlsafe_b64encode
from typing import Union
from urllib.parse import urlencode

import nest_asyncio
from httpx import AsyncClient, Request
from loguru import logger
from oscrypto.asymmetric import generate_pair

from pullkin.proto.android_checkin_proto import AndroidCheckinProto, ChromeBuildProto
from pullkin.proto.checkin_proto import AndroidCheckinRequest, AndroidCheckinResponse
from pullkin.proto.mcs_proto import *

nest_asyncio.apply()
logger.disable("pullkin")


class PullkinBase:
    unicode = str

    SERVER_KEY = (
        b"\x04\x33\x94\xf7\xdf\xa1\xeb\xb1\xdc\x03\xa2\x5e\x15\x71\xdb\x48\xd3"
        b"\x2e\xed\xed\xb2\x34\xdb\xb7\x47\x3a\x0c\x8f\xc4\xcc\xe1\x6f\x3c"
        b"\x8c\x84\xdf\xab\xb6\x66\x3e\xf2\x0c\xd4\x8b\xfe\xe3\xf9\x76\x2f"
        b"\x14\x1c\x63\x08\x6a\x6f\x2d\xb1\x1a\x95\xb0\xce\x37\xc0\x9c\x6e"
    )

    REGISTER_URL = "https://android.clients.google.com/c2dm/register3"
    CHECKIN_URL = "https://android.clients.google.com/checkin"
    FCM_SUBSCRIBE = "https://fcm.googleapis.com/fcm/connect/subscribe"
    FCM_ENDPOINT = "https://fcm.googleapis.com/fcm/send"

    PUSH_HOST = "mtalk.google.com"
    PUSH_PORT = 5228

    MCS_VERSION = 41

    PACKET_BY_TAG = [
        HeartbeatPing,
        HeartbeatAck,
        LoginRequest,
        LoginResponse,
        Close,
        "MessageStanza",
        "PresenceStanza",
        IqStanza,
        DataMessageStanza,
        "BatchPresenceStanza",
        StreamErrorStanza,
        "HttpRequest",
        "HttpResponse",
        "BindAccountRequest",
        "BindAccountResponse",
        "TalkMetadata",
    ]
    packet_union = Union[
        HeartbeatPing,
        HeartbeatAck,
        LoginRequest,
        LoginResponse,
        Close,
        IqStanza,
        DataMessageStanza,
        StreamErrorStanza,
    ]

    def __init__(self):
        self._http_client = None

    @property
    def http_client(self):
        if not self._http_client:
            self._http_client = AsyncClient()
        return self._http_client

    async def close(self):
        if self._http_client:
            await self._http_client.aclose()
            self._http_client = None

    async def _do_request(self, req, retries=5):
        for _ in range(retries):
            try:
                resp = await self.http_client.send(req, follow_redirects=True)
                resp_data = resp.content
                logger.debug(f"Response:\n{resp_data}")
                return resp_data
            except Exception as e:
                logger.opt(exception=e).debug("Error during request:")
                time.sleep(1)
        raise ConnectionError(f"Error during request: {e}")

    async def gcm_check_in(self, androidId=None, securityToken=None, **_):
        """
        perform check-in request

        androidId, securityToken can be provided if we already did the initial
        check-in

        returns dict with androidId, securityToken and more
        """
        chrome = ChromeBuildProto()
        chrome.platform = 3
        chrome.chrome_version = "63.0.3234.0"
        chrome.channel = 1

        checkin = AndroidCheckinProto()
        checkin.type = 3
        checkin.chrome_build.from_dict(chrome.to_dict())

        payload = AndroidCheckinRequest()
        payload.user_serial_number = 0
        payload.checkin.from_dict(checkin.to_dict())
        payload.version = 3
        if androidId:
            payload.id = int(androidId)
        if securityToken:
            payload.security_token = int(securityToken)

        logger.debug(f"Payload:\n{payload}")
        req = self.http_client.build_request(
            method="POST",
            url=self.CHECKIN_URL,
            headers={"Content-Type": "application/x-protobuf"},
            content=payload.SerializeToString(),
            timeout=5
        )
        resp_data = await self._do_request(req)
        resp = AndroidCheckinResponse()
        resp.parse(resp_data)
        logger.debug(f"Response:\n{resp}")
        return resp.to_dict()

    @classmethod
    def urlsafe_base64(cls, data):
        """
        base64-encodes data with -_ instead of +/ and removes all = padding.
        also strips newlines

        returns a string
        """
        res = urlsafe_b64encode(data).replace(b"=", b"")
        return res.replace(b"\n", b"").decode("ascii")

    async def gcm_register(self, appId, retries=5, **_):
        """
        obtains a gcm token

        appId: app id as an integer
        retries: number of failed requests before giving up

        returns {"token": "...", "appId": 123123, "androidId":123123,
                 "securityToken": 123123}
        """
        # contains androidId, securityToken and more
        chk = await self.gcm_check_in()
        logger.debug(f"Check_in:\n{chk}")
        body = {
            "app": "org.chromium.linux",
            "X-subtype": appId,
            "device": chk["androidId"],
            "sender": self.urlsafe_base64(self.SERVER_KEY),
        }
        data = urlencode(body)
        logger.debug(f"Data:\n{data}")
        auth = "AidLogin {}:{}".format(chk["androidId"], chk["securityToken"])
        req = Request(
            method="POST",
            url=self.REGISTER_URL,
            headers={"Authorization": auth},
            data=body,
        )
        for _ in range(retries):
            resp_data = await self._do_request(req, retries)
            if b"Error" in resp_data:
                err = resp_data.decode("utf-8")
                logger.error(f"Register request has failed with {err}")
                continue
            token = resp_data.decode("utf-8").split("=")[1]
            chkfields = {k: chk[k] for k in ["androidId", "securityToken"]}
            res = {"token": token, "appId": appId}
            res.update(chkfields)
            return res
        return None

    async def fcm_register(self, sender_id, token, retries=5):
        """
        generates key pair and obtains a fcm token

        sender_id: sender id as an integer
        token: the subscription token in the dict returned by gcm_register

        returns {"keys": keys, "fcm": {...}}
        """
        # I used this analyzer to figure out how to slice the asn1 structs
        # https://lapo.it/asn1js
        # first byte of public key is skipped for some reason
        # maybe it's always zero
        public, private = generate_pair("ec", curve=self.unicode("secp256r1"))
        from base64 import b64encode

        logger.debug("# Public")
        logger.debug(b64encode(public.asn1.dump()))
        logger.debug("# Private")
        logger.debug(b64encode(private.asn1.dump()))
        keys = {
            "public": self.urlsafe_base64(public.asn1.dump()[26:]),
            "private": self.urlsafe_base64(private.asn1.dump()),
            "secret": self.urlsafe_base64(os.urandom(16)),
        }
        body = {
            "authorized_entity": sender_id,
            "endpoint": "{}/{}".format(self.FCM_ENDPOINT, token),
            "encryption_key": keys["public"],
            "encryption_auth": keys["secret"],
        }
        logger.debug(f"Data:\n{body}")
        req = Request(method="POST", url=self.FCM_SUBSCRIBE, data=body)
        resp_data = await self._do_request(req, retries)
        return {"keys": keys, "fcm": json.loads(resp_data.decode("utf-8"))}

    async def _register(self, sender_id):
        """register gcm and fcm tokens for sender_id"""
        app_id = "1:302251869498:android:90c5cd74bae68792813c03"
        subscription = await self.gcm_register(appId=app_id)
        logger.debug(f"GCM subscription data: {subscription}")
        fcm = await self.fcm_register(sender_id=sender_id, token=subscription["token"])
        logger.debug(f"FCM subscription data: {fcm}")
        res = {"gcm": subscription}
        res.update(fcm)
        return res

    def register(self, sender_id):
        """
        Sync version. Register "app" for receive pushed

        Returns "app"-credential in dict for receive "personal" push by token
        """
        res = asyncio.get_event_loop().run_until_complete(self._register(sender_id))
        return res

    @classmethod
    def _encode_varint32(cls, x):
        res = bytearray([])
        while x != 0:
            b = x & 0x7F
            x >>= 7
            if x != 0:
                b |= 0x80
            res.append(b)
        return bytes(res)

    @classmethod
    def _app_data_by_key(cls, p, key, blow_shit_up=True):
        for x in p.app_data:
            if x.key == key:
                return x.value
        if blow_shit_up:
            raise RuntimeError("couldn't find in app_data {}".format(key))
        return None

    @classmethod
    def _is_deleted_message(cls, p):
        for x in p.app_data:
            if x.key == "message_type" and x.value == "deleted_messages":
                return True
        return False
